# -*- coding: utf-8 -*-
"""
Created on Wed Jan 22 10:07:12 2025

@author: mstep
"""

#%%
import math


def follow_path(path, distance):
    '''
    Follow the points in path for the given distance.
    Return the points visited.
    Update path to continue following its points.
    '''
    from_point = path.pop(0)
    segment = [from_point]
    while len(path) > 0 and distance > 0:
        # Move from from_point toward path[0].
        to_point = path[0]
        segment_length = math.dist(from_point, to_point)

        # See how far it is to path[0].
        if segment_length >= distance:
            # It's too far. Only move part of the way there.
            dx = to_point[0] - from_point[0]
            dy = to_point[1] - from_point[1]
            nx = dx / segment_length
            ny = dy / segment_length
            stop_point = (
                from_point[0] + nx * distance,
                from_point[1] + ny * distance
            )
            segment.append(stop_point)

            # If stop_point is not close to path[0],
            # start the next segment from stop_point.
            if not math.isclose(segment_length, distance):
                path.insert(0, stop_point)

            # Return.
            return segment
        else:
            # We can reach path[0]. Move there.
            segment.append(to_point)
            distance -= segment_length
            from_point = to_point
            path.pop(0)

    # If we get here, we either ran out of path or distance.
    # Return whatever is left.
    return segment

def get_dashed_segments(path, dashes):
    '''Follow the points in path. The dashes parameter holds draw/skip values.'''
    segments = []
    dash_index = 0
    while len(path) > 0:
        # Draw a segment.
        segment = follow_path(path, dashes[dash_index])
        segments.append(segment)
        if len(path) == 0: break
        dash_index = (dash_index + 1) % len(dashes)

        # Skip a segment.
        segment = follow_path(path, dashes[dash_index])
        dash_index = (dash_index + 1) % len(dashes)

    return segments

#%%
def draw_dashed_polygon(dr, points, outline=None, fill=None, thickness=1,
        joint='miter', dashes=(5, 5)):
    '''Draw a dashed polygon.'''
    # Copy the points list so we don't mess up the original.
    points = points.copy()
    points = points + [points[0]]  # Close the polygon.

    # Add a tiny bit of the next line segment to round the final corner.
    dx = points[1][0] - points[0][0]
    dy = points[1][1] - points[0][1]
    dist = math.dist(points[0], points[1])
    dx = dx / dist * 0.000001
    dy = dy / dist * 0.000001
    points.append((points[0][0] + dx, points[0][1] + dy))

    # Draw the polyline points.
    draw_dashed_polyline(dr, points, outline, fill, thickness, joint, dashes)

def draw_dashed_polyline(dr, points, outline=None, fill=None, thickness=1,
        joint='miter', dashes=(5, 5)):
    '''Draw a dashed polyline.'''
    # If we should fill the shape, do so.
    if fill is not None:
        dr.polygon(points, fill=fill, width=thickness)

    # Copy the points so we don't mess up the original list.
    path = points.copy()

    # Draw the outline.
    for segment in get_dashed_segments(path, dashes):
        dr.line(segment, fill=outline, width=thickness, joint=joint)

#%%

def center_text(dr, cx, cy, text, font, text_fill='black', rect_outline=None):
    '''Draw the text centered at (cx, cy).'''
    # dr.textbbox returns x, y, width, height.
    left, top, wid, hgt = dr.textbbox((0, 0), text, font=font)
    x = cx - wid / 2
    y = cy - hgt / 2
    dr.text((x, y), text, font=font, fill=text_fill)

    if rect_outline is not None:
        dr.rectangle((x, y, x + wid, y + hgt), outline=rect_outline)

#%%
import tkinter as tk
from PIL import Image, ImageDraw, ImageTk, ImageFont

class DashedLinesApp:
    def __init__(self):
        self.window = tk.Tk()

        self.window.title('pil_dashed_lines')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('300x300')

        # Create the background image.
        wid = 570
        hgt = 560
        result_image = Image.new("RGB", (wid, hgt), (255, 255, 255))

        # Create a Draw object to draw on the image.
        result_dr = ImageDraw.Draw(result_image)

        # Draw some dashed shapes.
        self.draw_shapes(result_dr)

        # Save the result_image (just to show we can).
        result_image.save('result.png')

        # Convert the image into a PhotoImage for tkinter to display.
        photo_image = ImageTk.PhotoImage(result_image)

        # Create the canvas sized to fit the image.
        self.canvas = tk.Canvas(self.window, bg='white',
            width=wid, height=hgt,
            borderwidth=0, highlightthickness=0)
        margin = 10
        self.canvas.pack(side=tk.TOP, anchor=tk.NW, padx=margin, pady=margin)

        # Display the image on the canvas.
        self.canvas.create_image(0, 0, image=photo_image, anchor=tk.NW) 

        # Size the window to fit.
        self.window.geometry(f'{wid + 2 * margin}x{hgt + 2 * margin}')

        # Display the window.
        self.window.focus_force()
        self.window.mainloop()

    def kill_callback(self):
        # Destroy the tkinter window.
        self.window.destroy()

    def draw_shapes(self, dr):
        '''Draw some dashed shapes.'''
        # Draw the shapes starting at (10, 10), thickness = 1, unfilled.
        self.draw_shapes_at(dr, 10, 10)

        # Draw the shapes starting at (300, 10), thickness = 10, unfilled.
        self.draw_shapes_at(dr, 300, 10, 10)

        # Draw the shapes starting at (300, 10), thickness = 10, filled.
        self.draw_shapes_at(dr, 10, 310, 10, 'lightblue', 'miter')

        # Draw the shapes starting at (300, 10), thickness = 10, filled.
        self.draw_shapes_at(dr, 300, 310, 10, 'lightgreen', 'curve')

    def draw_shapes_at(self, dr, x0, y0, thickness=1, fill=None, joint=None):
        '''Draw some dashed shapes.'''
        # Long line.
        x = x0
        y = y0
        points = [(x, y), (x + 250, y + 20)]
        dashes = (10, 10)
        draw_dashed_polyline(dr, points, outline='red', dashes=dashes,
            thickness=thickness, fill=fill, joint=joint)
        # Joint types: miter, curve, None

        # Long line with complex dash pattern.
        y += 20
        points = [(x, y), (x + 250, y + 20)]
        dashes = (20, 10, 5)
        draw_dashed_polyline(dr, points, outline='black', dashes=dashes,
            thickness=thickness, fill=fill, joint=joint)
        # Joint types: miter, curve, None

        # Zig zag.
        y += 30
        delta = 20
        points = []
        for i in range(7):
            points.append((x, y))
            x += delta
            y += delta
            points.append((x, y))
            x += delta
            y -= delta
        dashes = (5, 5)
        draw_dashed_polyline(dr, points, outline='green', dashes=dashes,
            thickness=thickness // 2, fill=fill, joint=joint)

        # Circle.
        radius = 40
        cx = x0 + radius
        cy = y + radius + 40
        points = []
        for angle in range(0, 360, 10):
            px = cx + radius * math.cos(math.radians(angle))
            py = cy + radius * math.sin(math.radians(angle))
            points.append((px, py))
        dashes = (5, 5)
        draw_dashed_polygon(dr, points, outline='blue', dashes=dashes,
            thickness=thickness // 2, fill=fill, joint=joint)

        # Triangle.
        side_length = 4 * radius / math.sqrt(3)
        dx = side_length / 2
        dy = dx * math.sqrt(3)
        x = cx + radius + 10 + dx
        y = cy - radius
        points = [(x, y), (x + dx, y + dy), (x - dx, y + dy)]
        circumference = 3 * side_length
        dash = circumference / 13
        dashes = (dash, dash)
        draw_dashed_polygon(dr, points, outline='orange', dashes=dashes,
            thickness=thickness, fill=fill, joint=joint)

        # Rectangle.
        x = x0
        y = cy + radius + 10
        wid = 250
        hgt = 50
        points = [(x, y), (x + wid, y), (x + wid, y + hgt), (x, y + hgt)]
        dash = 2 * (wid + hgt) / 24.7
        dashes = (dash, dash)
        draw_dashed_polygon(dr, points, outline='purple', dashes=dashes,
            thickness=thickness, fill=fill, joint=joint)

        # Draw the joint name.
        cx = x + wid / 2
        cy = y + hgt / 2
        text = 'None' if joint is None else joint
        text = 'Joint = ' + text
        font = ImageFont.truetype("arial.ttf", 20)
        center_text(dr, cx, cy, text, font, text_fill='blue',
            rect_outline=None)

#%%
DashedLinesApp()
print('Done')
